#!~/virtualenv-1.9/myEV/bin/python
# encoding=utf-8

""" Functions for querying DBpedia """

from subprocess import Popen, PIPE


def executeQuery(q):
    """ Execute query q over local DBpedia storage (Virtuoso)
    """
    # TODO odstranit heslo ze zdrojaku
    q = 'SPARQL %s;' % q
    proc = Popen(['isql-vt', 'localhost', 'dba', 'n3jakadatabaz3'], \
            stdout=PIPE, stderr=PIPE, stdin=PIPE)
    result, err = proc.communicate(input=q)
    # isql-vt outputs a lot of text around the answer, which we will remove
    results = [line for line in result.split('\n')[8:-3]]
    return results


def findAbstract(keyword):
    """ Find an abstract for given keyword.
    """
    def isRelevant(abstract):
        return len(abstract) > 15 and 'přesměruj' not in abstract

    keyword = keyword.replace(' ', '_')
    #query = """
    #    PREFIX csrsrc: <http://cs.dbpedia.org/resource/>
    #    PREFIX dbo: <http://dbpedia.org/ontology/> 
    #    SELECT ?abstract                                                    
    #    WHERE {                                                                     
    #        csrsrc:%s dbo:abstract ?abstract .
    #    }""" % keyword
    query = """
        PREFIX csrsrc: <http://cs.dbpedia.org/resource/>
        SELECT ?c                                                   
        WHERE {                                                                     
            <http://cs.dbpedia.org/resource/%s>
                <http://www.w3.org/2000/01/rdf-schema#comment> ?c .
        }""" % keyword
    result = executeQuery(query)
    return result[0] if len(result) > 0 and isRelevant(result[0]) else None


# TODO opravit, aby fungovalo - zrejme udelat nejake mapovani znamych a
# pouzitelnych properties (mapovani muze obsahovat i schema otazky pro danou
# property
def findProperties(keyword):
    # prilis obecne, nefunguje spravne
    keyword = keyword.replace(' ', '_')
    query4 = """
        PREFIX csrsrc: <http://cs.dbpedia.org/resource/>
        PREFIX dbo: <http://dbpedia.org/ontology/> 
        SELECT ?p ?o                                                    
        WHERE {                                                                     
            csrsrc:%s ?p ?o .
        }""" % keyword
    query = """
        PREFIX csrsrc: <http://cs.dbpedia.org/resource/>
        PREFIX csprop: <http://cs.dbpedia.org/property/>
        SELECT ?l ?o                                                   
        WHERE {                                                                     
            csrsrc:%s ?p ?o .
            ?p <http://www.w3.org/1999/02/22-rdf-syntax-ns#type>
                <http://www.w3.org/1999/02/22-rdf-syntax-ns#Property> .
            ?p <http://www.w3.org/2000/01/rdf-schema#label> ?l .
        }""" % keyword
    results = executeQuery(query)
    #results = ['prop' + r for r in results]
    results = [r.split(None, 1) for r in results]
    return results


def findAllFacts(keyword):
    """ Find all facts about given keyword in DBpediea.
    """
    keyword = keyword.replace(' ', '_')
    query = """
        PREFIX csrsrc: <http://cs.dbpedia.org/resource/>
        SELECT ?p ?o                                                    
        WHERE {                                                                     
            <http://cs.dbpedia.org/resource/%s> ?p ?o .
        }""" % keyword
    results = executeQuery(query)
    results = [r.split(None, 1) for r in results]
    return results


# TODO vytvorit vertikal kategorii indikujicich osobu (pododbne jako places?)
# load a set of place categories
# TODO nahradit absolutni cestu relativni, ale tak, aby to fungovalo 
typeCategories = {
        'Place': '/home/xeffenb1/procvicnik/data/placeCategories.vert',
        'Person': '/home/xeffenb1/procvicnik/data/personCategories.vert'
        }
# replace files by actual set of categories
for t, vertFile in typeCategories.items():
    with open(vertFile) as f:
        categorySet = set(f.read().split('\n'))
    typeCategories[t] = categorySet

def findTypesAndCategories(keyword):
    """ Find categories which given keyword belongs to.
    """
    def inferTypes(categories):
        inferredTypes = []
        for t in typeCategories:
            for c in categories:
                if c in typeCategories[t]:
                    inferredTypes.append(t)
                    break
        return inferredTypes
    
    if isinstance(keyword, unicode):
        keyword = keyword.encode('utf-8')
    keyword = keyword.replace(' ', '_')
    # prefixy v dotazech nejde pohodlne pouzit kvuli jmenum s teckou
    # (napr. Vaclav IV.)
    queryCategories = """
        PREFIX csrsrc: <http://cs.dbpedia.org/resource/>
        SELECT ?label                                                    
        WHERE {                                                                     
            <http://cs.dbpedia.org/resource/%s>
                <http://purl.org/dc/terms/subject> ?c .
            ?c <http://www.w3.org/2000/01/rdf-schema#label> ?label
        }"""  % keyword
    queryTypes = """
        PREFIX csrsrc: <http://cs.dbpedia.org/resource/>
        SELECT ?c                                                    
        WHERE {                                                                     
            <http://cs.dbpedia.org/resource/%s>
                <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> ?o .
            ?o <http://www.w3.org/2000/01/rdf-schema#label> ?c
        }""" % keyword
    categories = executeQuery(queryCategories)
    types = executeQuery(queryTypes)
    types.extend(inferTypes(categories))
    # filtering 
    categories = [r for r  in categories if r != 'Muži' and r != keyword]
    return types, categories


def findSubcategories(categoryURI, deep=1, outputFile=None):
    """ Find subcategories of given category.
        Example of correct categoryURI:
            http://cs.dbpedia.org/resource/Kategorie:Kontinenty
    """

    def findSubcategoriesRec(categoryURI, deep, exclude=set()):
        if deep == 0:
            return set()
        query = """
            PREFIX csrsrc: <http://cs.dbpedia.org/resource/>                        
            PREFIX skos: <http://www.w3.org/2004/02/skos/core#>                       
            SELECT ?c                                                        
            WHERE {                                                                     
                ?c skos:broader <%s> .                                  
            }""" % categoryURI
        subcategories = set(executeQuery(query))
        subcategories.difference_update(exclude)

        subSubcategories = set()
        exclude = subcategories | exclude
        for s in subcategories:
            subSubcategories.update(findSubcategoriesRec(s, deep-1, exclude))
        return subcategories | subSubcategories

    subcategories = findSubcategoriesRec(categoryURI, deep=deep)
    labelQuery = """
            PREFIX csrsrc: <http://cs.dbpedia.org/resource/>
            SELECT ?label                                                    
            WHERE {                                                                     
                <%s> <http://www.w3.org/2000/01/rdf-schema#label> ?label
            }"""
    subcategories = [executeQuery(labelQuery % c)[0] for c in subcategories]
    if outputFile:
        with open(outputFile, 'w') as f:
            f.write('\n'.join(subcategories))
    return subcategories

def findPlaceCategories(outputFile='./data/placeCategories.vert'):
    PLACE_ROOT_CATEGORY = \
            "http://cs.dbpedia.org/resource/Kategorie:Členění_území"
    categories = findSubcategories(PLACE_ROOT_CATEGORY, deep = 3)
    categories = filter(lambda x: 'podle zem' not in x, categories)
    with open(outputFile, 'w') as f:
        f.write('\n'.join(categories))

    

if __name__ == '__main__':
    import json
    print json.dumps(findAllFacts('Bitva u Lipan'), indent=2)
