#!~/virtualenv-1.9/myEV/bin/python
# encoding=utf-8

""" Functions for analyses and processing text. """

from __future__ import unicode_literals
import sys
import re
from subprocess import Popen, PIPE
sys.path.append('/nlp/projekty/declension')                                       
import declension
import random

#MAJKA = "/nlp/projekty/ajka/bin/majka"
SET = "/nlp/projekty/set/set/set.py"
UNITOK = "/nlp/corpora/programy/unitok.py"
DESAMB = "/nlp/corpora/programy/desamb.utf8.majka.sh"
#sys.path.append('/nlp/projekty/set/set')
#import set as SET



def tokenize(text):
    """ Transform text into list of tokens.
    """
    if type(text) == unicode:
        text = text.encode('utf-8')
    proc = Popen([UNITOK, '--language=czech'], stdout=PIPE, \
        stderr=PIPE, stdin=PIPE)
    tokenizedText, err = proc.communicate(input=text)
    # TODO vyresit kodovani (jiz zde prevadet na unicode?)
    return tokenizedText


def lemmatize(text, isTokenized=False):
    """ Find lemmas and morfological tags for given text.
    """
    tokenizedText = text if isTokenized else tokenize(text)
    # TODO volat desamb z /nlp/.... a svoji kopii smazat
    proc = Popen([DESAMB, '-', '1'], stdout=PIPE, \
        stderr=PIPE, stdin=PIPE)
    lemmatizedText, err = proc.communicate(input=tokenizedText)
    return unicode(lemmatizedText, 'utf-8')


def createLemmaDict(phrases):
    """ Find lemmas of given list of phrases.
    """
    text = str(' | ').join(phrases)
    text, romanNumbers = extractRomanNumbers(text)
    lemmatizedText = [t.split() for t in lemmatize(text).split('\n')]
    lemmatizedText = expandTokens(lemmatizedText, romanNumbers)
    lemmata = []
    lemma = []
    for parts in lemmatizedText:
        if parts[0] == '|':
            lemmata.append(lemma)
            lemma = []
        elif len(parts) == 2 and parts[1].startswith('hack'):
            lemma = lemma[:-1] # desamb obcas zaludne prida tecky...
        elif len(parts) == 3:
            if parts[1] == '#num#':
                lemma.append(parts[0])
            else:
                lemma.append(parts[1])
        elif parts[0] == '<g/>':
            lemma.append(parts[0])
    if lemma:
        lemmata.append(lemma)
    lemmaDict = {}
    if len(lemmata) != len(phrases):
        raise Exception('Number of lemmas and phrases have to be equal.')
    for lemma, phrase in zip(lemmata, phrases):
        lemmaDict[tuple(lemma)] = phrase
    return lemmaDict
    
def firstColumn(text):                                                          
    """ Return only first column from multicolumn text.
    """
    # TODO prepsat bez volani externiho cutu
    #proc = Popen(['cut', '-f1'], stdout=PIPE, stderr=PIPE, stdin=PIPE)          
    #return proc.communicate(input=text)[0]  
    lines = text.split('\n')
    splitLines = [l.split(None, 1) for l in lines]
    return [l[0] for l in splitLines if len(l) > 0] 


#def getSentences(text):
#    """ Divide given text into sentences. """
#    # Kdy nefunguje spravne: tecky v datumech a jmenech (Karel IV.)
#    # TODO opravit ^
#    tokens = firstColumn(lemmatize(text))
#    sentences = []
#    for token in tokens:
#        if token.startswith("<s"):
#            sentence = ""
#        elif token == "</s>":
#            sentences.append(sentence[:-1])
#        elif token == "<g/>":
#            sentence = sentence[:-1]
#        else:
#            sentence += token + " "
#    # provizorni filtrovani vet, ktere se nepodarilo rozdelit
#    # + odstraneni kontextovych vet (TODO anafory)
#    return [s for s in sentences if len(s) < 220 and isContextFree(s)]


def glueLemmatizedText(tokens):
    glueText = ""
    for token in tokens:
        if token[0] == "<g/>":
            glueText = glueText[:-1]
        elif len(token) >= 3: 
            glueText += token[0] + ' '
    return glueText.strip()


def syntacticAnalysis(text):
    """ Syntactic analysis of text using SET.
    """
    vert = lemmatize(tokenize(text))
    proc = Popen([SET, "-d"], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    output, err = proc.communicate(input=vert)
    print 'OUTPUT:\n', output


#CONTEXT_SENTENCE_RE = re.compile(r"""
#          (?: .* [\s,;:])?
#          (?: ten | ta | to | tento | tato | toto | toho | tomu | tu | té
#                | této | tomto | jiný | jiné | jiné | další | také | tak
#                | jeho | její | jejího | obě | oběma 
#                | až | už | předtím | potom) [\s,;:]
#        """, re.VERBOSE | re.IGNORECASE)

CONTEXT_WORDS = {'ten', 'ta', 'to' 'tento', 'tato', 'toto', 'toho', \
            'tomu', 'tu', 'té', 'této', 'jiný', 'jiné', 'jiná', 'další',
            'také', 'téhož', 'tak', 'těchto', 'tomto', \
            'jeho', 'její', 'jejího', 'jejích', 'jich', 'jejich', \
            'němu', 'ní', \
            'obě', 'oběma', 'avšak', \
            'už', 'předtím', 'potom', 'poté',\
            '[', 'kam', 'odkud', 'aby'}
def isContextFree(sentence):
    """ Return True if the given sentence is meaningful without context.
    """
    if sentence[-1][0] == '?':
        return False
    # prvni slovo je sloveso
    if sentence[0][2][1] == '5':
        return False
    verbCount = 0
    nounAppeared = False
    verbDistance = 10
    for i, token in enumerate(sentence):
        if token[0].lower() in CONTEXT_WORDS:
            return False
        verbDistance += 1
        # TODO pozor na Tuple Index Out of Range
        if len(token) <= 2: print token
        if token[2][1] == '5' \
                and token[2][7] != 'F': # sloveso (ne infinitiv)
            if not nounAppeared:
                # verb before noun
                return False
            # podvod: pokud je pred slovesem 'a', vyhodime z vety jeji zbytek,
            # a tim zachranime jeji bezkontextovost TODO udelat to
            # transparentneji
            if i > 6 and verbCount > 0 and sentence[i-1][0] \
                    in ['a','i','ale','ovšem','nicméně','avšak']:
                end = i - 2
                while sentence[end][0] in ['<g/>',',',';',':']:
                    end -= 1
                del sentence[end+1:len(sentence)-2]
                return True
            if verbDistance > 3:
                verbCount += 1
            verbDistance = 0
        elif token[2][1] == '1' or len(token) == 4:
            nounAppeared = True
    if verbCount > 2 or verbCount == 0:
        return False
    return True

#def shallowParsing(sentence):
#    """ Return list of (word, lemma, tag)
#    """
#    tokens = lemmatize(sentence)
#    words = []
#    for token in tokens.split('\n'):
#        if token == "<g/>":
#            words.append(("<g/>","", ""))    
#        else:
#            parts = token.split()
#            if len(parts) == 3:
#                if not parts[1].istitle():
#                    parts[0] = parts[0].lower()
#                words.append(tuple(parts))
#    return words

def sentenceToPhrase(sentence):
    """ Lower first letter of the first word (if not a name)
        and remove dot at the end of the sentence.
    """
    # TODO osetrit degenerovane vety (musi obsahovat alespon 1 slovo!)
    if sentence[-1][0] in '.?!':
        if sentence[-2][0] == '<g/>':
            sentence = sentence[:-2]
        else:
            sentence = sentence[:-1]
    else:
        sentence = sentence[:]
    if len(sentence[0]) == 3 and not sentence[0][1].istitle():
        # we don't want to change the original token
        # (is's still in the original sentence)
        sentence[0] = (sentence[0][0].lower(), sentence[0][1], sentence[0][2])
    return sentence


def isVerb(word):
    """ Return True if given word is a verb.
    """
    # TODO pro slova vytvorit jednuchou tridu
    if len(word) < 2:
        return False
    tag = word[2]
    return len(tag) >= 2 and tag[1] == '5'


def hasType(token, kwType):
    """ Return true if the token is a keyword of given type (kwType)
    """
    return len(token) == 4 and token[3] == kwType

def isKeyword(token):
    return len(token) == 4


def declineNounPhrase(np, case):
    # declension.py doesn't handle lower/upper cases correctly
    firstLetters = [w[0] for w in np.split()]
    declinedNP = declension.main(['c1', 'c'+case, '', np.encode('utf-8')])[0]
    # declension vraci unicode/string podle toho, zda vstupni retezec
    # obsahoval/neobsahoval ne-ascii zanky!
    if type(declinedNP)==str:
        declinedNP = unicode(declinedNP, 'utf-8')
    declinedNP = ' '.join([(w if w[0] == l else w.capitalize()) \
            for l, w in zip(firstLetters, declinedNP.split())])
    return declinedNP

#def createDeclensionRegex(word):
#    # TODO ...
#    print declension.main(['c1', 'c2', '', 'Severní Amerika'.encode('utf-8')])

# regular expression for date
PART_OF_YEAR = r"""(?:
    v\ zimě | na\ jaře | na\ podzim | v\ létě
        |   (?:dne[ ])?
            \d{1,2}\.
            \s*
            (?: \d{1,2}\. | ledna | února | března | dubna | května | června
                | července | srpna | září | října | listopadu | prosince)
    )"""
YEAR_RE = re.compile(r"""
        (?:
            (?: již\ | už\ )?
            (?: %(PART_OF_YEAR)s[ ])?
            (?: (?:z\ | na\ počátku\ | na\ konci\ | počátkem\ )?
                    roku | v\ roce)
          | %(PART_OF_YEAR)s
        )
        [ ](\d+) 
    """ % {'PART_OF_YEAR': PART_OF_YEAR}, re.VERBOSE | re.IGNORECASE)

def extractDates(text):
    """ Search for dates in text, return simplified text and all dates.
    """
    text = text.replace('DNES', 'Dnes')
    dates = YEAR_RE.findall(text)
    text = YEAR_RE.sub(' DNES ', text)
    #for dates = [m.  YEAR_RE.finditer(text):
    #    print m.span(), m.group(), m.group(1)    
    return text, dates

# regular expression for roman number followed by a dot
ROMAN_RE = re.compile(r"""
        ( M{0,4}
            (?: CM|CD|D?C{0,3} )
            (?: XC|XL|L?X{0,3} )
            (?: IX|IV|V?I{1,3} )
          | M{0,4}
            (?: CM|CD|D?C{0,3} )
            (?: XC|XL|L?X{1,3} )
          | M{0,4}
            (?: CM|CD|D?C{1,3} )
          | M{1,4}
        )
        \.
        """, re.VERBOSE )

def extractRomanNumbers(text):
    """ Search for roman numbers in text, return simplified text and all
        roman numbers.
    """
    text = text.replace('NUMBERTH', 'Numberth') # just for sure
    romanNumbers = ROMAN_RE.findall(text)
    text = ROMAN_RE.sub('NUMBERTH', text)
    return text, romanNumbers


def expandTokens(tokens, romanNumbers):
    expTokens = []
    romanIndex = 0
    for t in tokens:
        if len(t) > 0:
            if t[0] == 'NUMBERTH':
                expTokens.append((romanNumbers[romanIndex], '#num#', 'k4'))
                expTokens.append(('<g/>',))
                expTokens.append(('.', '.', 'kIx.'))
                romanIndex += 1
            else:
                expTokens.append(tuple(t))
    # TODO opravit
    #if romanIndex != len(romanNumbers):
    #    raise Exception('Bad number of roman numbers.')
    return expTokens

def guessNounPhraseTag(tokens):
    """ Return tag of first noun in the phrase.
        If there is not any noun, return tag of first word in the phrase.
    """
    for t in tokens:
        if len(t) == 3 and t[2][1] == '1': # noun
            return t[2]
    return tokens[0][2]

def parseText(text, keywords, dates, people, places):
    """ Divide text into sentences, tag each word and label keywords.
    """
    def match(splitKeyword, tokens, i):
        """ Return True if the keyword starts on the i-th index
            in list of tokens
        """
        for k, w in enumerate(splitKeyword):
            if len(tokens[i+k]) > 1 and tokens[i+k][1] != '#num#':
                if tokens[i+k][1] != splitKeyword[k]:
                    return False
            else:
                if tokens[i+k][0] != splitKeyword[k]:
                    return False
        return True

    text, romanNumbers = extractRomanNumbers(text)

    #print text
    #print 'dates:'
    #print dates
    #print keywords
    # find lemmata of all words in text
    tokens = [t.split() for t in lemmatize(text).split('\n')]
    tokens = expandTokens(tokens, romanNumbers)

    # create bigram dictionary
    bigrams = createBigramsDict(tokens)

    # find lemmata of keywords
    lemmaDict = createLemmaDict(keywords)

    keywordsOnPosition = {}
    for keywordLemma in lemmaDict:
        for i in range(len(tokens) - len(keywordLemma)):
            if match(keywordLemma, tokens, i):
                if i not in keywordsOnPosition \
                        or len(keywordLemma) > len(keywordsOnPosition[i]):
                    keywordsOnPosition[i] = keywordLemma
   
    sentences = []
    ignoreNext = 0
    dateIndex = 0
    for i, token in enumerate(tokens):
        if token[0] == 'DNES':
            sentence.append(('v roce '+dates[dateIndex], dates[dateIndex], \
                    'k4', 'Date'))
            dateIndex += 1
        elif ignoreNext > 0:
            ignoreNext -= 1
            continue
        elif i in keywordsOnPosition:
            keywordLemma = keywordsOnPosition[i]
            l = len(keywordLemma)
            phrase = lemmaDict[keywordLemma]
            originalText = glueLemmatizedText(tokens[i:i+l])
            tag = guessNounPhraseTag(tokens[i:i+l])
            if phrase in people:
                kwType = 'Person'
            elif phrase in places:
                kwType = 'Place'
            else:
                kwType = 'kw'
            sentence.append((originalText, phrase, tag, kwType))
            ignoreNext = l - 1
        elif token[0].startswith("<s"):
            sentence = []
        elif token[0] == "</s>":
            sentences.append(sentence)
        elif token[0].startswith('<g'):
            sentence.append(('<g/>', '', '--'))
        else:
            sentence.append(token)

    if dateIndex != len(dates):
        print 
        print dateIndex
        print dates
        raise Exception('Bad number of dates.')

    # provizorni filtrovani vet, ktere se nepodarilo rozdelit
    # + odstraneni kontextovych vet (TODO anafory)
    sentences = [s for s in sentences if len(s) < 50 and isContextFree(s)]
    return sentences, bigrams


#def formWhenQuestion(event):
#    for i, word in enumerate(event):
#        if isVerb(word) or word[0] == "se":
#            break
#    else:
#        return "Kdy nastala tato udalost: " + glueLemmatizedText(event) + "?"
#
#    question = 'Kdy %s %s?' % (word[0], \
#            glueLemmatizedText(event[:i] + event[i+1:]))
#    return question

def createBigramsDict(tokens):
    def isWord(t):
        return not t.startswith('<') and t != 'DNES' and t not in '.,;:?!()'

    bigrams = {'start':['42'], 'Protože':['je', 'není', 'měl']}
    # TODO ma smysl brat v uvahu zacatky a konce vet pro vetsi prirozenost
    for i in range(len(tokens)-1):
        w1 = tokens[i]
        w2 = tokens[i+1]
        if isWord(w1[0]) and isWord(w2[0]):
            bigrams.setdefault(w1[0], [])
            bigrams[w1[0]].append(w2[0])
            if w1[2][1] == '1':
                bigrams['start'].append(w1[0])
    return bigrams

def formWQuestion(tokens, questionWord):
    for i, word in enumerate(tokens):
        # TODO rozlisovat se jako zajmeno a predlozku nebo tak neco
        if isVerb(word) or word[0] == "se":
            break
    else:
        return "Určete %s %s" % (questionWord, glueLemmatizedText(tokens))

    question = '%s %s %s?' % (questionWord.capitalize(), word[0], \
            glueLemmatizedText(tokens[:i] + tokens[i+1:]))
    return question


def sentenceFromPhrase(phrase, beginning='', end=''):
    sentence = '%s%s%s' % (beginning, glueLemmatizedText(phrase), end)
    return sentence[0].upper() + sentence[1:]

#def formWhyQuestion(fact, reason):
#    question = 'Proč %s?' % glueLemmatizedText(fact)
#    answer = 'Protože %s.' % glueLemmatizedText(reason)
#    return question, answer

# TODO vety by si zaslouzili vlastni tridu
def printSentence(s):
    for token in s:
        form = '{:23} '*(len(token)-1) + '{}'
        print form.format(*token)
    print '------------------------------'


def phraseGenerator(bigrams, beginning='', end=''):
    text = beginning
    word = beginning if beginning else 'start'
    for i in range(random.randint(3,5)):
        if word not in bigrams:
            word = 'start'
        nextWord = random.choice(bigrams[word])
        text += " " + nextWord
        word = nextWord
    return text.lstrip() + end

def randomNameGenerator():
    # TODO z nejakeho seznamu jmen, nebo permutace casti jmen v textu
    # nebo ze slabik vytvorit zcela nove jmeno
    return random.choice(['Pepa z Depa', 'Karel IV.', 'Superman', \
            'Ash Ketchum', 'Salman Khan', 'Alois Jirásek'])

def randomPlaceGenerator():
    # TODO ze seznamu mist nebo generovat zcela fiktivni jmena mist
    return random.choice(['Brno', 'Austrálie', 'Dolní Lhota', 'Aljaška', \
            'Portoriko', 'Kent', 'Krkonoše'])

