#!~/virtualenv-1.9/myEV/bin/python
# encoding=utf-8

""" Module for creating ontology from html page. """

from __future__ import unicode_literals
import sys
# provizorne kvuli problemu s virtualenv
sys.path.append('/usr/lib/python2.7/dist-packages')
import re
from subprocess import Popen, PIPE
from language_tools import extractDates, parseText, hasType, isKeyword,\
        sentenceToPhrase, printSentence
from dbpedia import findTypesAndCategories, findAbstract
from html_tools import getHtml, getInformationContent
from rdflib import Graph, Literal, URIRef, BNode, Namespace
from rdflib.namespace import RDF, RDFS, FOAF
from translations import findGlobalURIs, findUsableFactsRemotely
#sys.path.append('/nlp/projekty/topicks/')                                       
#import topic3
import json

from language_tools import glueLemmatizedText

# exercise namespace
EXERCISES = Namespace('http://effaacademy.cz/exercises/')
KEYWORD = Namespace('http://effaacademy.cz/exercises/keyword/')


def createOntology(url):
    htmlPage = getHtml(url)
    isWiki = ('wikipedia' in url)
    return createOntologyFromText(htmlPage, isWiki)


TEST = False
def createOntologyFromText(htmlPage, isWiki=False, translations=False):
    """ Return ontology of the given html page represented using triples
        in RDF graph.
    """
    ontology = Graph()
  
    if not TEST:
        text, keywords, images = getInformationContent(htmlPage, isWiki)
    else:
        with open('data/mishmash.txt') as f:
            text = f.read()
        with open('data/mishmash.kw') as f:
            keywords = json.load(f)
        images = []
    
    # bind our namespace with prefix to make querying easier
    ontology.bind('exrc', EXERCISES)

    # decode to unicode
    text = unicode(text, 'utf-8') if isinstance(text, str) else text

    if not isWiki: # nebo testovat spis pocet nalezenych keywords?
        print 'topics start'
        topicksKeywords = findTopicsInPlainText(text)
        print 'topics end'
        keywords.extend(topicksKeywords)

    keywords = [(unicode(k, 'utf-8') if isinstance(k, str) else k) \
            for k in keywords if len(k) > 1]
   
    # extract dates from text
    text, dates = extractDates(text)
    for d in dates:
        dateNode = KEYWORD.term(d)
        ontology.add( (dateNode, EXERCISES.type, Literal('Date')) )
        ontology.add( (dateNode, RDFS.label, Literal(d)) )
    # TODO ? remove dates from keywords ?

    people = set()
    places = set()
    for keyword in keywords:
        # TODO pojmy, ktere jsme nalezli pres odkazy by bylo lepsi
        # identifikovat pomoci jejich URI
        keywordNode = KEYWORD.term(keyword.replace(' ', '_'))
        ontology.add( (keywordNode, RDF.type, EXERCISES.keyword) ) 
        ontology.add( (keywordNode, RDFS.label, Literal(keyword)) )
        #abstract = findAbstract(keyword) # TODO pozor na kodovani
        #if abstract:
        #    ontology.add((keywordNode, EXERCISES.abstract, Literal(abstract)))
        types, categories = findTypesAndCategories(keyword)
        for category in categories: 
            category = category.decode('utf-8')
            if not spamCategory(category):
                ontology.add((keywordNode, EXERCISES.category, Literal(category)))
        for t in types:
            t = t.decode('utf-8')
            if t != 'A thing' and t != 'Thing': 
                ontology.add( (keywordNode, EXERCISES.type, Literal(t)) )
                # TODO napsat obecneji
                if t == 'Person':
                    people.add(keyword)
                elif t == 'Place':
                    places.add(keyword)
    
    # simple heuristic to estimate topic of the text
    # TODO vzit v uvahu nadpis (neni vzdy primocare)
    peopleRate = len(people) + 1
    placesRate = len(places) + 1
    datesRate = len(dates) + 1
    total = peopleRate + placesRate + datesRate + 3
    peopleRate = float(peopleRate)/total
    datesRate = float(datesRate)/total
    placesRate = float(placesRate)/total
    ontology.add( (EXERCISES.topic, EXERCISES.history,
        Literal(str(datesRate))) )
    ontology.add( (EXERCISES.topic, EXERCISES.geography,
        Literal(str(placesRate))) )
    ontology.add( (EXERCISES.topic, EXERCISES.people,
        Literal(str(peopleRate))) )


    for (caption, src) in images:
        imageNode = BNode()
        ontology.add((imageNode, RDF.type, EXERCISES.image))
        ontology.add((imageNode, RDFS.label, Literal(caption)))
        # TODO vybrat konzistentni a kompatibilni ontologii !!
        ontology.add((imageNode, EXERCISES.url, Literal(src)))

    sentences, bigrams = parseText(text, keywords, dates, people, places)
    #for s in sentences:
    #    print s
    #for b in bigrams:
    #    print b
    #    print bigrams[b]

    MATCH_FUNCTIONS = [matchDateEvent, matchPlaceFact, matchHowManyFact, \
            matchExampleFact, matchPersonFact, matchAnyFact]
    for s in sentences:
        for matchFunction in MATCH_FUNCTIONS:
            fact = matchFunction(s)
            if fact:
                factLabel, factType, factKeyInfo = fact
                factLabel = json.dumps(factLabel)
                factKeyInfo = json.dumps(factKeyInfo)
                factNode = BNode()
                ontology.add( (factNode, RDF.type, EXERCISES.fact) )
                ontology.add( (factNode, RDFS.label, Literal(factLabel)) )
                ontology.add( (factNode, EXERCISES[factType],
                    Literal(factKeyInfo)) )

    # why questions (reasons) handled separetedly, because the match function
    # needs previous sentence
    for i, s in enumerate(sentences):
        prevSentence = sentences[i-1] if i > 0 else []
        reasonFact = matchReasonFact(prevSentence, s)
        if reasonFact:
            factLabel, factType, reason = reasonFact
            factLabel = json.dumps(factLabel)
            reason = json.dumps(reason)
            factNode = BNode()
            ontology.add((factNode, RDF.type, EXERCISES.fact))
            ontology.add((factNode, RDFS.label, Literal(factLabel)))
            ontology.add((factNode, EXERCISES[factType], Literal(reason)) )

    if translations:
        expandOntologyUsingTranslations(ontology, keywords, \
                people, places, dates)

    if TEST:
        with open('/home/xeffenb1/procvicnik/data/mishmash.ontl', 'w') as f:
            f.write(ontology.serialize(format='turtle'))

    return ontology, bigrams # TODO bigramy ?? 
    #return ontology

# TODO pravdepodonbe potreba rozdelit text do mensich casti, aby to topisc
# zvladl; dale vyzkouset omezeni na pocet vracenych slov (vliv na rychlost)
def findTopicsInPlainText(text):
    """ Use topics to extract keywords from plain text.
    """
    # topics obcas skonci se Segmentation fault !!!
    proc = Popen(['python', 'findTopicks.py'], \
            stdout=PIPE, stderr=PIPE, stdin=PIPE)
    results, err = proc.communicate(input=text[:7000].encode('utf-8'))
    if err or not results:
        #print 'error occured'
        return []
    else:
        results = results.decode('utf-8')
        results = json.loads(results)
        return results


def matchDateEvent(sentence):
    """ Returns name of event and year when the event happened from the given
        sentence. If there is no year in sentence, returns None.
    """
    datesIndexes = []
    for i, token in enumerate(sentence):
        if hasType(token, 'Date'):
            datesIndexes.append(i)
    if len(datesIndexes) == 1:
        i = datesIndexes[0]
        eventLabel = sentenceToPhrase(sentence[:i] + sentence[i+1:])
        return eventLabel, 'when', sentence[i],
   

def matchPersonFact(sentence):
    tokens = sentenceToPhrase(sentence)
    if hasType(tokens[0], 'Person'):
        return tokens[1:], 'who', tokens[0]
    
def matchHowManyFact(sentence):
    numberIdx = [i for i in range(len(sentence)) if sentence[i][1] == '#num#']
    if 1 <= len(numberIdx) <= 2:
        i = numberIdx[0]
        if i+3 < len(sentence) and sentence[i+1][0] == '<g/>': return
        if i > 0 and sentence[i-1][0] == '<g/>': return
        fact = sentence[:i] + [('kolik','','')] + sentence[i+1:-1]
        return fact, 'howMany', sentence[i]

def matchExampleFact(sentence):
    for i, token in enumerate(sentence):
        if token[0] in ['např.', 'například'] and i < len(sentence) \
                and isKeyword(sentence[i+1]):
            example = sentence[i+1]
            while i>3 and sentence[i-1][0] in ['<g/>',',','(']:
                i -= 1
            if i > 3:
                return sentence[:i], 'example', example


def matchPlaceFact(sentence):
    tokens = sentenceToPhrase(sentence)
    if hasType(tokens[-1], 'Place') and tokens[-3][0] != 'a' \
            and len(tokens) < 15:
        if tokens[-2][0] in ['v', 've', 'na']:
            return tokens[:-2], 'where', tokens[-1]
        elif tokens[-2][0] in ['do']:
            return tokens[:-2], 'whereTo', tokens[-1]
        elif tokens[-2][0] in ['z', 'ze', 'od']:
            return tokens[:-2], 'whereFrom', tokens[-1]
    elif hasType(tokens[0], 'Place'):
        return tokens[1:], 'whichPlace', tokens[0]
    
    
def matchAnyFact(sentence):
    for token in reversed(sentence):
        if isKeyword(token):
            return sentence, 'includesKeyword', token

    # TODO nejobecnejsi varianta ('Doplnte') obecneji i pro osoby...
    #else:
    #    answer = [sentence[i][1] for i in placeIndexes]
    #    sentence = [(('???', '???', '--') if hasType(t, 'Place') else t) \
    #            for t in sentence]
    #    print 'Doplňte: ' + glueLemmatizedText(sentence)
    #    print ', '.join(answer)

    #print '*****************'
    #raw_input()


# TODO osetrit pripad, kdy je predchozi veta vypustena jako kontextova
def matchReasonFact(prevSentence, sentence):
    sentence = sentenceToPhrase(sentence)
    if sentence[0][0] in ['proto', 'tudíž']:
        prevSentence = sentenceToPhrase(prevSentence)
        return sentence[1:], 'why', prevSentence
        #return formWhyQuestion(sentence[1:], prevSentence)
    elif (sentence[0][0],sentence[1][0]) in [('a','proto'), ('a','tudíž')]:
        prevSentence = sentenceToPhrase(prevSentence)
        return sentence[2:], 'why', prevSentence
        #return formWhyQuestion(sentence[2:], prevSentence)
    else:
        for i, token in enumerate(sentence):
            if token[0] in ['proto', 'tudíž']:
                k = i - 1
                while sentence[k][0] in ['<g/>', 'a', ',']:
                    k -= 1
                fact = sentence[i+1:]
                reason = sentence[:k+1]
                return fact, 'why', reason
            elif token[0] in ['protože', 'neboť']:
                k = i - 1
                while sentence[k][0] in ['<g/>', 'a', ',']:
                    k -= 1
                fact = sentence[:k+1]
                reason = sentence[i+1:]
                return fact, 'why', reason

def findAll(typeLabel, ontology):
    labels = ontology.query(
        """SELECT DISTINCT ?label
            WHERE {
                ?node exrc:type "%s" .
                ?node rdfs:label ?label .
            }""" % typeLabel)
    return [unicode(l[0]) for l in labels]


def findAllKeywords(ontology):
    labels = ontology.query(
        """SELECT DISTINCT ?label
            WHERE {
                ?node a exrc:keyword .
                ?node rdfs:label ?label .
            }""")
    return [unicode(l[0]) for l in labels]


def checkType(keyword, kwType, ontology):
    """ Return True if the keyword has type kwType under the ontology.
    """
    types = ontology.objects(subject=KEYWORD.keyword, predicate=RDF.type)
    return kwType in types


SPAM_CATEGORIES = set(['Leden', 'Únor', 'Březen', 'Duben', 'Květen', \
            'Červen', 'Červenec', 'Srpen', 'Září', 'Říjen', 'Listopad', \
            'Prosinec'])
def spamCategory(category):
    return category in SPAM_CATEGORIES


def expandOntologyUsingTranslations(ontology, keywords, people, places, dates):
    for keyword in keywords:
        uriDict = findGlobalURIs(keyword)
        if 'dbpedia' in uriDict:
            facts = findUsableFactsRemotely(uriDict['dbpedia'])
            for (pred, obj) in facts:
                keywordNode = KEYWORD.term(keyword.replace(' ', '_'))
                # TODO ne vzdy to musi byt literal (ale mozna to nevadi)
                ontology.add( (keywordNode, EXERCISES[pred], Literal(obj)) )

