#!~/virtualenv-1.9/myEV/bin/python
# encoding=utf-8

""" Searching for facts in english DBpedia using translation of a term.  """

from __future__ import unicode_literals
from rdflib import Graph, URIRef
from dbpedia import executeQuery
from SPARQLWrapper import SPARQLWrapper, JSON
import json

GLOBAL_NS = [
    ('dbpedia', 'http://dbpedia.org/resource/'),
    ('freebase', 'http://rdf.freebase.com/ns/'),
    ('wikidata', 'http://www.wikidata.org/entity/')]
def findGlobalURIs(term):
    ''' Find global (english) identifiers of given term. '''
    term = term.replace(' ', '_')
    query = """
        PREFIX csrsrc: <http://cs.dbpedia.org/resource/>
        SELECT ?uri                                               
        WHERE {                                                                     
            <http://cs.dbpedia.org/resource/%s>
            <http://www.w3.org/2002/07/owl#sameAs> ?uri .
        }""" % term
    results = executeQuery(query.encode('utf-8'))
    uriDict = {}
    for r in results:
        r = unicode(r, 'utf-8')
        for label, ns in GLOBAL_NS:
            if r.startswith(ns):
                uriDict[label] = r
    return uriDict


def findAllFactsRemotely(keyword):
    """ Find all facts about given keyword in public dbpedia.org endpoint
    """
    keyword = keyword.replace(' ', '_')
    uriDict = findGlobalURIs(keyword)
    g = Graph()
    if 'dbpedia' in uriDict:
        g.parse(uriDict['dbpedia'])

    print g.serialize(format='turtle')
    #return results


def remoteQuery(query):
    sparql = SPARQLWrapper("http://dbpedia.org/sparql")
    sparql.setQuery(query)
    sparql.setReturnFormat(JSON)
    results = sparql.query().convert()
    return results


def remoteSubjectsQuery(query):
    results = remoteQuery(query)
    #print json.dumps(results, indent=1)
    #return [(p, results['results']['bindings'][0][p]['value']) \
    #        for p in results['head']['vars']]
    results = results['results']['bindings']
    return [(p,o['value']) for r in results for (p, o) in r.items()] 


def findPersonFactsRemotely(dbpediaUri):
    """ Find facts about given person through public DBpedia endpoint.
    """
    query = """
        PREFIX dbpedia-owl: <http://dbpedia.org/ontology/>
        SELECT ?birthYear, ?deathYear, ?image, ?type, ?bp
        WHERE {                                                                     
            {<%(uri)s> dbpedia-owl:birthYear ?birthYear}
            UNION {<%(uri)s> dbpedia-owl:deathDate ?deathYear}
            UNION {<%(uri)s> dbpedia-owl:thumbnail ?image}
            #UNION {<%(uri)s> dbpedia-owl:birthPlace ?bp}
            #UNION {<%(uri)s> a ?type}
            #UNION {<%(uri)s dcterms:subject ?category>}
        }""" % {'uri':dbpediaUri}
    facts = remoteSubjectsQuery(query)
    #for p, o in facts:
    #    print p, ':', o
    return facts


def findUsableFactsRemotely(dbpediaUri):
    """ Find usable facts about given keyword through public DBpedia endpoint.
    """
    query = """
        PREFIX dbpedia-owl: <http://dbpedia.org/ontology/>
        PREFIX geo: <http://www.w3.org/2003/01/geo/wgs84_pos#>
        SELECT ?type, ?birthYear, ?deathYear, ?image, ?date,
            ?latitude, ?longitude
        WHERE {
            {<%(uri)s> a ?type}                                                     
            #UNION {<%(uri)s dcterms:subject ?category>}
            UNION {<%(uri)s> dbpedia-owl:birthYear ?birthYear}
            UNION {<%(uri)s> dbpedia-owl:birthDate ?birthYear}
            UNION {<%(uri)s> dbpedia-owl:deathYear ?deathYear}
            UNION {<%(uri)s> dbpedia-owl:deathDate ?deathYear}
            #UNION {<%(uri)s> dbpedia-owl:birthPlace ?bp}
            UNION {<%(uri)s> dbpedia-owl:thumbnail ?image}
            UNION {<%(uri)s> dbpedia-owl:date ?date}
            UNION {<%(uri)s> geo:lat ?latitude}
            UNION {<%(uri)s> geo:long ?longitude}
        }""" % {'uri':dbpediaUri}
    facts = remoteSubjectsQuery(query)
    return facts


# zajimava fakta:
# typy a kategorie -> predikat 'a', prip. wiki categories (dcterms:subject),
# dbpprop:wordnet_type
# dalsi veci (muze se lisit Event/Osoba,Person)
#  dbpedia-owl:date :place, :result, :combatant 
# geo souradnice grs:point nebo geo:geometry nebo geo:lat, geo:lon
# osoby: dbpedia-owl:birthYear :deathDate, birthPlace, deathPlace, era
#        fofa:depiction nebo thumbnail


if __name__ == '__main__':
    keyword = 'Bitva u Lipan'
    #keyword = 'Jan Hus'
    uriDict = findGlobalURIs(keyword)
    print json.dumps(findUsableFactsRemotely(uriDict['dbpedia']), indent=1)
    #findAllFactsRemotely(keyword)
