#!~/virtualenv-1.9/myEV/bin/python                                              
# encoding=utf-8

from __future__ import unicode_literals
import sys
sys.path.insert(0,'/home/xeffenb1/procvicnik')
from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
from SocketServer import ThreadingMixIn
from jinja2 import Template
from urlparse import urlparse, parse_qs
from ontology import createOntology
from exercises import createExercises
from collections import OrderedDict
import json
import logging

PORT = 4197
TEMPLATE_DIR = 'templates/'
LOG_DIRECTORY = '/nlp/projekty/procvicnik/logs/'
WWW_DIR = '.'

def renderTemplate(templateFileName, **kwargs):
    ''' Render html template.
    '''
    with open(TEMPLATE_DIR + templateFileName) as f:
        template = Template(f.read().decode('utf-8'))
    return template.render(kwargs)


def createOntologyRequest(params):
    # TODO tohle je provizorni reseni -> rozdelit do fazi pro plynulost
    # (napred ontoogie -> ulozit do DB, pak postupne otazky...)
    ontology, bigrams = createOntology(params['url'])
    exercises = createExercises(ontology, bigrams)
    #return [{'priority':p,'exrc':e.__dict__} for (p,e) in exercises]
    return [e.__dict__ for e in exercises] # TODO rozsiritelneji
    #return [{'question':'Kde...?', 'choices':['A','B','C'], \
    #        'correctAnswer': 2}, {'question':'Kdy...?', 'choices':['d','e','f'], \
    #        'correctAnswer': 1}]
  

EXERCISES_TYPES = OrderedDict([
        ('who', 'Kdo? Komu? Koho?'),
        ('where', 'Kde? Odkud? Kam?'),
        ('when', 'Kdy?'),
        ('howMany', 'Kolik?'),
        ('why', 'Proč?'),
        ('example', 'Uveďte příklad.'),
        ('image', 'Co je na obrázku?'),
        ('category', 'Kdo/co nepatří mezi ostatní?'),
        ('gap', 'Doplňte ...'),
        ('couldMeet', 'Mohli se potkat?'),
        ('eventOnMap', 'Najděte na mapě.')])

def exampleExerciseRequest(params):
    exerciseType = params['exercise-type']
    if exerciseType in EXERCISES_TYPES.keys():
        exampleExercisesFile = 'exercises-examples/%s.json' % exerciseType
        with open(exampleExercisesFile) as f:
            exercises = json.load(f)
            return exercises
    else:
        return {'status':'Failed. Given exercise type does not exist.'}


def exerciseDoneRequest(params):
    logExercise(params['exercise'])
    # TODO return updated list of exercises (some of them)
    return {}


STATIC_FILES = [
        ('.jpg', 'image/jpg'),
        ('.png', 'image/png'),
        ('.gif', 'image/gif'),
        ('.ico', 'image/ico'),
        ('.css', 'text/css'),
        ('.js', 'application/javascript')]

REQUEST_INTERFACE = [
        ('/interface/create-ontology', createOntologyRequest),
        ('/interface/exercise-done', exerciseDoneRequest),
        ('/interface/example', exampleExerciseRequest)
        ]

class MyRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        """ Respond to a GET request.
        """
        path = urlparse(self.path).path
        params = parse_qs(urlparse(self.path).query)

        # serve static files
        for ending, minitype in STATIC_FILES:
            if path.endswith(ending):
                try:
                    with open(WWW_DIR + path) as f:
                        fileContent = f.read()
                except IOError:
                    self.send_error(404, 'File Not Found: %s' % path)
                    return
                self.send_response(200)
                self.send_header('Content-type', minitype)
                self.end_headers()
                self.wfile.write(fileContent)
                return

        if path.startswith('/example/'):
            exrcType = path[9:]
            content = renderTemplate('exercise.html', url='', example=exrcType)
        elif 'url' in params:
            # TODO lepsi zpusob predani url javascriptu na strane klienta
            content = renderTemplate('exercise.html', url=params['url'][0])
        else:
            content = renderTemplate('main_page.html',
                    exercisesTypes=EXERCISES_TYPES)
        wholePage = renderTemplate('template.html', content=content)

        # send response
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        self.wfile.write(wholePage.encode('utf-8'))

    def do_POST(self):
        """ Respond to a POST request.
        """
        path = urlparse(self.path).path
        length = int(self.headers['Content-Length'])
        params = json.loads(self.rfile.read(length).decode('utf-8'))
        #fs = cgi.FieldStorage(
        #        fp=self.rfile,
        #        headers=self.headers,
        #        environ={"REQUEST_METHOD":"POST"})
        #params = {k:fs.getvalue(k) for k in fs.keys()}

        # serve predefined requests
        for requestPath, respondFunction in REQUEST_INTERFACE:
            if path == requestPath:
                result = json.dumps(respondFunction(params))
                self.send_response(200)
                self.send_header('Content-type', 'application/json')
                self.end_headers()
                self.wfile.write(result.encode('utf-8'))
                return


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """ Server that handles requests in separates threads.
    """

FEEDBACK_FLAGS = ['nonsence', 'grammarError', 'nonrelevant', 'difficult', \
        'easy', 'badAnswers', 'great']
def logExercise(exrc):
    choices = '||'.join(exrc['choices']) # TODO neni 100%, doladit
    feedback = ''.join([str(int(exrc[x])) for x in FEEDBACK_FLAGS])
    if exrc['answer'] == -1 and not '1' in feedback:
        return
    logText = '\n'.join([
        str(exrc['timestamp']),
        exrc['url'],
        exrc['question'],
        choices,
        str(exrc['correctAnswer']),
        str(exrc['answer']),
        feedback])
    # TODO logovat do DB
    logging.info(logText)

def main():
    # logging setting
    # TODO pojmenovany logger atd. (viz. logging tutorial)
    logging.basicConfig(
            filename=LOG_DIRECTORY+'exercises.log',
            level=logging.INFO,
            format='%(message)s')

    # start server
    server = ThreadedHTTPServer(('', PORT), MyRequestHandler)
    print 'Server is serving at port %s, use ^C to stop.' % str(PORT)
    server.serve_forever()


if __name__ == '__main__':
    main()
